import abc
from typing import Callable, Dict, Optional

import distrax as dx
import jax.numpy as jnp
import tensorflow_probability.substrates.jax.bijectors as tfb
from chex import dataclass

from gpjax.config import get_defaults

from .config import Identity, Softplus, add_parameter
from .gps import Posterior
from .types import Array, Dataset
from .utils import I, concat_dictionaries

DEFAULT_JITTER = get_defaults()["jitter"]

Diagonal = dx.Lambda(
    forward=lambda x: jnp.diagflat(x), inverse=lambda x: jnp.diagonal(x)
)

FillDiagonal = dx.Chain([Diagonal, Softplus])
FillTriangular = dx.Chain([tfb.FillTriangular()])


@dataclass
class VariationalFamily:
    @property
    @abc.abstractmethod
    def params(self) -> Dict:
        raise NotImplementedError


@dataclass
class VariationalGaussian(VariationalFamily):
    inducing_inputs: Array
    name: str = "Gaussian"
    variational_mean: Optional[Array] = None
    variational_root_covariance: Optional[Array] = None
    diag: Optional[bool] = False
    whiten: Optional[bool] = True

    def __post_init__(self):
        """Initialize the variational family."""
        self.num_inducing = self.inducing_inputs.shape[0]
        add_parameter("inducing_inputs", Identity)

        nz = self.num_inducing

        if self.variational_mean is None:
            self.variational_mean = jnp.zeros((nz, 1))
            add_parameter("variational_mean", Identity)

        if self.variational_root_covariance is None:
            self.variational_root_covariance = I(nz)
            if self.diag:
                add_parameter("variational_root_covariance", FillDiagonal)
            else:
                add_parameter("variational_root_covariance", FillTriangular)

    @property
    def params(self) -> Dict:
        hyperparams = {
            "inducing_inputs": self.inducing_inputs,
            "variational_mean": self.variational_mean,
            "variational_root_covariance": self.variational_root_covariance,
        }
        return hyperparams



import typing as tp

import jax
import optax
import tensorflow.data as tfd
from tqdm import trange

from .parameters import trainable_params
from .types import Dataset


def fit(
    objective: tp.Callable,
    params: tp.Dict,
    trainables: tp.Dict,
    opt_init,
    opt_update,
    get_params,
    n_iters: int = 100,
    log_rate: int = 10,
    jit_compile: bool = False,
) -> tp.Dict:
    """Abstracted method for fitting a GP model with respect to a supplied objective function.
    Optimisers used here should originate from Jax's experimental module.
    Args:
        objective (tp.Callable): The objective function that we are optimising with respect to.
        params (dict): The parameters for which we would like to minimise our objective function wtih.
        trainables (dict): Boolean dictionary of same structure as 'params' that determines which parameters should be trained.
        opt_init (tp.Callable): The supplied optimiser's initialisation function.
        opt_update (tp.Callable): Optimiser's update method.
        get_params (tp.Callable): Return the current parameter state set from the optimiser.
        n_iters (int, optional): The number of optimisation steps to run. Defaults to 100.
        log_rate (int, optional): How frequently the objective function's value should be printed. Defaults to 10.
        jit_compile (bool,  optional): Jit compiles the loss function and training step within the training loop.
    Returns:
        tp.Dict: An optimised set of parameters.
    """
    opt_state = opt_init(params)

    def loss(params):
        params = trainable_params(params, trainables)
        return objective(params)

    def step(i, opt_state):
        params = get_params(opt_state)
        loss_val, loss_gradient = jax.value_and_grad(loss)(params)
        return opt_update(i, loss_gradient, opt_state), loss_val

    if jit_compile:
        loss = jax.jit(loss)
        step = jax.jit(step)

    tr = trange(n_iters)
    for i in tr:
        opt_state, val = step(i, opt_state)
        if i % log_rate == 0 or i == n_iters:
            tr.set_postfix({"Objective": f"{val: .2f}"})
    return get_params(opt_state)


def optax_fit(
    objective: tp.Callable,
    params: tp.Dict,
    trainables: tp.Dict,
    optax_optim,
    n_iters: int = 100,
    log_rate: int = 10,
    jit_compile: bool = False,
) -> tp.Dict:
    """Abstracted method for fitting a GP model with respect to a supplied objective function.
    Optimisers used here should originate from Optax.
    Args:
        objective (tp.Callable): The objective function that we are optimising with respect to.
        params (dict): The parameters for which we would like to minimise our objective function wtih.
        trainables (dict): Boolean dictionary of same structure as 'params' that determines which parameters should be trained.
        optax_optim (GradientTransformation): The Optax optimiser that is to be used for learning a parameter set.
        n_iters (int, optional): The number of optimisation steps to run. Defaults to 100.
        log_rate (int, optional): How frequently the objective function's value should be printed. Defaults to 10.
        jit_compile (bool,  optional): Jit compiles the loss function and training step within the training loop.
    Returns:
        tp.Dict: An optimised set of parameters.
    """
    opt_state = optax_optim.init(params)

    def loss(params):
        params = trainable_params(params, trainables)
        return objective(params)

    def step(params, opt_state):
        loss_val, loss_gradient = jax.value_and_grad(loss)(params)
        updates, opt_state = optax_optim.update(loss_gradient, opt_state, params)
        params = optax.apply_updates(params, updates)
        return params, opt_state, loss_val

    if jit_compile:
        loss = jax.jit(loss)
        step = jax.jit(step)

    tr = trange(n_iters)
    for i in tr:
        params, opt_state, val = step(params, opt_state)
        if i % log_rate == 0 or i == n_iters:
            tr.set_postfix({"Objective": f"{val: .2f}"})
    return params


# Mini-batch loader from a TensorFlow Dataset:
def batch_loader(data: tfd.Dataset) -> tp.Callable[[None], Dataset]:
    """Abstracted method for loading mini-batches from a TensorFlow Dataset as a GPJax Dataset.
    Args:
        dataset (tfd.Dataset): Training data as a TensorFlow Dataset.
    Returns:
        tp.Callable[[None], Dataset]: A function that produces mini-batches as a GPJax Dataset.
    """

    dataset_iter = iter(data)

    def next_batch() -> Dataset:
        x_batch, y_batch = dataset_iter.next()
        return Dataset(X=x_batch.numpy(), y=y_batch.numpy())

    return next_batch


# Mini-batch gradient descent:
def fit_batches(
    objective: tp.Callable,
    params: tp.Dict,
    trainables: tp.Dict,
    opt_init,
    opt_update,
    get_params,
    get_batch,
    n_iters: tp.Optional[int] = 100,
    log_rate: tp.Optional[int] = 10,
    jit_compile: bool = False,
) -> tp.Dict:
    """Abstracted method for fitting a GP model with mini-batches respect to a supplied objective function.
    Optimisers used here should originate from Jax's experimental module.
    Args:
        objective (tp.Callable): The objective function that we are optimising with respect to.
        params (dict): The parameters for which we would like to minimise our objective function wtih.
        trainables (dict): Boolean dictionary of same structure as 'params' that determines which parameters should be trained.
        opt_init (tp.Callable): The supplied optimiser's initialisation function.
        opt_update (tp.Callable): Optimiser's update method.
        get_params (tp.Callable): Return the current parameter state set from the optimiser.
        n_iters (int, optional): The number of optimisation steps to run. Defaults to 100.
        log_rate (int, optional): How frequently the objective function's value should be printed. Defaults to 10.
        jit_compile (bool,  optional): Jit compiles the loss function and training step within the training loop.
    Returns:
        tp.Dict: An optimised set of parameters.
    """

    opt_state = opt_init(params)

    def loss(params, batch):
        params = trainable_params(params, trainables)
        return objective(params, batch)

    def train_step(i, opt_state, batch):
        params = get_params(opt_state)
        loss_val, loss_gradient = jax.value_and_grad(loss)(params, batch)
        return opt_update(i, loss_gradient, opt_state), loss_val

    if jit_compile:
        loss = jax.jit(loss)
        train_step = jax.jit(train_step)

    tr = trange(n_iters)
    for i in tr:
        batch = get_batch()
        opt_state, val = train_step(i, opt_state, batch)
        if i % log_rate == 0 or i == n_iters:
            tr.set_postfix({"Objective": f"{val: .2f}"})
    return get_params(opt_state)



import typing as tp
from abc import abstractmethod, abstractproperty

import distrax as dx
import jax.numpy as jnp
from chex import dataclass
from jax.scipy.linalg import cho_factor, cho_solve, solve_triangular

from .config import get_defaults
from .kernels import Kernel, cross_covariance, gram
from .likelihoods import (
    Gaussian,
    Likelihood,
    NonConjugateLikelihoods,
    NonConjugateLikelihoodType,
)
from .mean_functions import MeanFunction, Zero
from .parameters import copy_dict_structure, evaluate_priors, transform
from .types import Array, Dataset
from .utils import I, concat_dictionaries

DEFAULT_JITTER = get_defaults()["jitter"]


@dataclass
class GP:
    """Abstract Gaussian process object."""

    @abstractmethod
    def mean(self) -> tp.Callable[[Dataset], Array]:
        """Compute the GP's mean function."""
        raise NotImplementedError

    @abstractmethod
    def variance(self) -> tp.Callable[[Dataset], Array]:
        """Compute the GP's variance function."""
        raise NotImplementedError

    @abstractproperty
    def params(self) -> tp.Dict:
        """Initialise the GP's parameter set"""
        raise NotImplementedError


#######################
# GP Priors
#######################
@dataclass(repr=False)
class Prior(GP):
    """A Gaussian process prior object. The GP is parameterised by a mean and kernel function."""

    kernel: Kernel
    mean_function: tp.Optional[MeanFunction] = Zero()
    name: tp.Optional[str] = "Prior"
    jitter: tp.Optional[float] = DEFAULT_JITTER

    def __mul__(self, other: Gaussian):
        """The product of a prior and likelihood is proportional to the posterior distribution. By computing the product of a GP prior and a likelihood object, a posterior GP object will be returned.
        Args:
            other (Likelihood): The likelihood distribution of the observed dataset.
        Returns:
            Posterior: The relevant GP posterior for the given prior and likelihood. Special cases are accounted for where the model is conjugate.
        """
        return construct_posterior(prior=self, likelihood=other)

    def __rmul__(self, other: Gaussian):
        """Reimplement the multiplication operator to allow for order-invariant product of a likelihood and a prior i.e., ."""
        return self.__mul__(other)

    def mean(self, params: dict) -> tp.Callable[[Array], Array]:
        """Compute the GP's prior mean function.
        Args:
            params (dict): The specific set of parameters for which the mean function should be defined for.
        Returns:
            tp.Callable[[Array], Array]: A mean function that accepts an input array for where the mean function should be evaluated at. The mean function's value at these points is then returned.
        """

        def mean_fn(test_inputs: Array):
            t = test_inputs
            mt = self.mean_function(t, params["mean_function"])
            return mt

        return mean_fn

    def variance(self, params: dict) -> tp.Callable[[Array], Array]:
        """Compute the GP's prior variance function.
        Args:
            params (dict): The specific set of parameters for which the variance function should be defined for.
        Returns:
            tp.Callable[[Array], Array]: A variance function that accepts an input array for where the variance function should be evaluated at. The variance function's value at these points is then returned as a covariance matrix.
        """

        def variance_fn(test_inputs: Array):
            t = test_inputs
            Ktt = gram(self.kernel, t, params["kernel"])
            return Ktt

        return variance_fn

    @property
    def params(self) -> dict:
        """Initialise the GP prior's parameter set"""
        return {
            "kernel": self.kernel.params,
            "mean_function": self.mean_function.params,
        }

    def random_variable(self, test_inputs: Array, params: dict) -> dx.Distribution:
        """Using the GP's mean and covariance functions, we can also construct the multivariate normal random variable.
        Args:
            test_points (Array): The points at which we'd like to evaluate our mean and covariance function.
            params (dict): The parameterisation of the GP prior for which the random variable should be computed for.
        Returns:
            dx.Distribution: A Distrax Multivariate Normal distribution.
        """
        t = test_inputs
        nt = t.shape[0]
        mt = self.mean(params)(t)
        Ktt = self.variance(params)(t)
        Ktt += I(nt) * self.jitter
        Lt = jnp.linalg.cholesky(Ktt)
        return dx.MultivariateNormalTri(jnp.atleast_1d(mt.squeeze()), Lt)


#######################
# GP Posteriors
#######################
@dataclass
class Posterior(GP):
    """The base GP posterior object conditioned on an observed dataset."""

    prior: Prior
    likelihood: Likelihood
    name: tp.Optional[str] = "GP Posterior"
    jitter: tp.Optional[float] = DEFAULT_JITTER

    @abstractmethod
    def mean(self, train_data: Dataset, params: dict) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @abstractmethod
    def variance(
        self, train_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        raise NotImplementedError

    @property
    def params(self) -> dict:
        return concat_dictionaries(
            self.prior.params, {"likelihood": self.likelihood.params}
        )


@dataclass
class ConjugatePosterior(Posterior):
    prior: Prior
    likelihood: Gaussian
    name: tp.Optional[str] = "ConjugatePosterior"
    jitter: tp.Optional[float] = DEFAULT_JITTER

    def mean(self, train_data: Dataset, params: dict) -> tp.Callable[[Array], Array]:
        x, y, n_data = train_data.X, train_data.y, train_data.n
        obs_noise = params["likelihood"]["obs_noise"]
        mx = self.prior.mean_function(x, params["mean_function"])

        # Precompute covariance matrices
        Kxx = gram(self.prior.kernel, x, params["kernel"])
        Kxx += I(n_data) * self.jitter
        Lx = cho_factor(Kxx + I(n_data) * obs_noise, lower=True)

        weights = cho_solve(Lx, y - mx)

        def mean_fn(test_inputs: Array) -> Array:
            t = test_inputs
            mt = self.prior.mean_function(t, params["mean_function"])
            Ktx = cross_covariance(self.prior.kernel, t, x, params["kernel"])
            return mt + jnp.dot(Ktx, weights)

        return mean_fn

    def variance(
        self, train_data: Dataset, params: dict
    ) -> tp.Callable[[Array], Array]:
        x, n_data = train_data.X, train_data.n
        obs_noise = params["likelihood"]["obs_noise"]
        Kxx = gram(self.prior.kernel, x, params["kernel"])
        Kxx += I(n_data) * self.jitter
        Lx = cho_factor(Kxx + I(n_data) * obs_noise, lower=True)

        def variance_fn(test_inputs: Array) -> Array:
            t = test_inputs
            Ktx = cross_covariance(self.prior.kernel, t, x, params["kernel"])
            Ktt = gram(self.prior.kernel, t, params["kernel"])
            latent_values = cho_solve(Lx, Ktx.T)
            return Ktt - jnp.dot(Ktx, latent_values)

        return variance_fn

    def marginal_log_likelihood(
        self,
        train_data: Dataset,
        transformations: tp.Dict,
        priors: dict = None,
        negative: bool = False,
    ) -> tp.Callable[[Dataset], Array]:
        x, y, n_data = train_data.X, train_data.y, train_data.n

        def mll(
            params: dict,
        ):
            params = transform(params=params, transform_map=transformations)

            obs_noise = params["likelihood"]["obs_noise"]
            mu = self.prior.mean_function(x, params)
            Kxx = gram(self.prior.kernel, x, params["kernel"])
            Kxx += I(n_data) * self.jitter
            Lx = jnp.linalg.cholesky(Kxx + I(n_data) * obs_noise)

            random_variable = dx.MultivariateNormalTri(mu.squeeze(), Lx)

            log_prior_density = evaluate_priors(params, priors)
            constant = jnp.array(-1.0) if negative else jnp.array(1.0)
            return constant * (
                random_variable.log_prob(y.squeeze()).mean() + log_prior_density
            )

        return mll


@dataclass
class NonConjugatePosterior(Posterior):
    prior: Prior
    likelihood: NonConjugateLikelihoodType
    name: tp.Optional[str] = "Non-Conjugate Posterior"
    jitter: tp.Optional[float] = DEFAULT_JITTER

    def __repr__(self):
        mean_fn_string = self.prior.mean_function.__repr__()
        kernel_string = self.prior.kernel.__repr__()
        likelihood_string = self.likelihood.__repr__()
        return (
            f"Non-Conjugate Posterior\n{'-'*80}\n- {mean_fn_string}\n-"
            f" {kernel_string}\n- {likelihood_string}"
        )

    @property
    def params(self) -> dict:
        hyperparameters = concat_dictionaries(
            self.prior.params, {"likelihood": self.likelihood.params}
        )
        hyperparameters["latent"] = jnp.zeros(shape=(self.likelihood.num_datapoints, 1))
        return hyperparameters

    def mean(self, train_data: Dataset, params: dict) -> tp.Callable[[Dataset], Array]:
        x, n_data = train_data.X, train_data.n
        Kxx = gram(self.prior.kernel, x, params["kernel"])
        Kxx += I(n_data) * self.jitter
        Lx = jnp.linalg.cholesky(Kxx)

        def mean_fn(test_inputs: Array) -> Array:
            t = test_inputs
            Ktx = cross_covariance(self.prior.kernel, t, x, params["kernel"])
            Ktt = gram(self.prior.kernel, t, params["kernel"])
            A = solve_triangular(Lx, Ktx.T, lower=True)
            latent_var = Ktt - jnp.sum(jnp.square(A), -2)
            latent_mean = jnp.matmul(A.T, params["latent"])

            lvar = jnp.diag(latent_var)

            moment_fn = self.likelihood.predictive_moment_fn
            pred_rv = moment_fn(latent_mean.ravel(), lvar)
            return pred_rv.mean().reshape(-1, 1)

        return mean_fn

    def variance(
        self, train_data: Dataset, params: dict
    ) -> tp.Callable[[Dataset], Array]:
        x, n_data = train_data.X, train_data.n
        Kxx = gram(self.prior.kernel, x, params["kernel"])
        Lx = jnp.linalg.cholesky(Kxx + I(n_data) * self.jitter)

        def variance_fn(test_inputs: Array) -> Array:
            t = test_inputs
            Ktx = cross_covariance(self.prior.kernel, t, x, params["kernel"])
            Ktt = gram(self.prior.kernel, t, params["kernel"])
            A = solve_triangular(Lx, Ktx.T, lower=True)
            latent_var = Ktt - jnp.sum(jnp.square(A), -2)
            latent_mean = jnp.matmul(A.T, params["latent"])

            lvar = jnp.diag(latent_var)

            moment_fn = self.likelihood.predictive_moment_fn
            pred_rv = moment_fn(latent_mean.ravel(), lvar)
            return jnp.diag(pred_rv.variance())

        return variance_fn

    def marginal_log_likelihood(
        self,
        train_data: Dataset,
        transformations: tp.Dict,
        priors: dict = None,
        negative: bool = False,
    ) -> tp.Callable[[Dataset], Array]:
        x, y, n_data = train_data.X, train_data.y, train_data.n

        if not priors:
            priors = copy_dict_structure(self.params)
            priors["latent"] = dx.Normal(loc=0.0, scale=1.0)

        def mll(params: dict):
            params = transform(params=params, transform_map=transformations)
            Kxx = gram(self.prior.kernel, x, params["kernel"])
            Kxx += I(n_data) * self.jitter
            Lx = jnp.linalg.cholesky(Kxx)
            Fx = jnp.matmul(Lx, params["latent"])
            rv = self.likelihood.link_function(Fx, params)
            ll = jnp.sum(rv.log_prob(y))

            log_prior_density = evaluate_priors(params, priors)
            constant = jnp.array(-1.0) if negative else jnp.array(1.0)
            return constant * (ll + log_prior_density)

        return mll


def construct_posterior(prior: Prior, likelihood: Likelihood) -> Posterior:
    if isinstance(likelihood, Gaussian):
        PosteriorGP = ConjugatePosterior
    elif any([isinstance(likelihood, l) for l in NonConjugateLikelihoods]):
        PosteriorGP = NonConjugatePosterior
    else:
        raise NotImplementedError(
            f"No posterior implemented for {likelihood.name} likelihood"
        )
    return PosteriorGP(prior=prior, likelihood=likelihood)



import abc
from typing import Callable, Optional

import distrax as dx
import jax.numpy as jnp
import jax.scipy as jsp
from chex import dataclass

from .types import Array


@dataclass(repr=False)
class Likelihood:
    num_datapoints: int  # The number of datapoints that the likelihood factorises over
    name: Optional[str] = "Likelihood"

    def __repr__(self):
        return f"{self.name} likelihood function"

    @property
    @abc.abstractmethod
    def params(self) -> dict:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def link_function(self) -> Callable:
        raise NotImplementedError


@dataclass(repr=False)
class Gaussian(Likelihood):
    name: Optional[str] = "Gaussian"

    @property
    def params(self) -> dict:
        return {"obs_noise": jnp.array([1.0])}

    @property
    def link_function(self) -> Callable:
        def link_fn(x, params: dict) -> dx.Distribution:
            return dx.Normal(loc=x, scale=params["obs_noise"])

        return link_fn


@dataclass(repr=False)
class Bernoulli(Likelihood):
    name: Optional[str] = "Bernoulli"

    @property
    def params(self) -> dict:
        return {}

    @property
    def link_function(self) -> Callable:
        def link_fn(x, params: dict) -> dx.Distribution:
            return dx.Bernoulli(probs=inv_probit(x))

        return link_fn

    @property
    def predictive_moment_fn(self) -> Callable:
        def moment_fn(mean: Array, variance: Array):
            rv = self.link_function(mean / jnp.sqrt(1 + variance), self.params)
            return rv

        return moment_fn


def inv_probit(x):
    jitter = 1e-3  # ensures output is strictly between 0 and 1
    return 0.5 * (1.0 + jsp.special.erf(x / jnp.sqrt(2.0))) * (1 - 2 * jitter) + jitter


NonConjugateLikelihoods = [Bernoulli]
NonConjugateLikelihoodType = Bernoulli  # Union[Bernoulli]



import abc
from typing import Callable, Dict, Optional, Tuple

import distrax as dx
import jax.numpy as jnp
from chex import dataclass
from jax import vmap
from jax.numpy.linalg import cholesky
from jax.scipy.linalg import solve_triangular

from .config import get_defaults
from .gps import Posterior
from .kernels import cross_covariance, gram
from .parameters import transform
from .quadrature import gauss_hermite_quadrature
from .types import Array, Dataset
from .utils import I, concat_dictionaries
from .variational import VariationalFamily

DEFAULT_JITTER = get_defaults()["jitter"]


@dataclass
class VariationalPosterior:
    posterior: Posterior
    variational_family: VariationalFamily
    jitter: Optional[float] = DEFAULT_JITTER

    def __post_init__(self):
        self.prior = self.posterior.prior
        self.likelihood = self.posterior.likelihood

    @property
    def params(self) -> Dict:
        hyperparams = concat_dictionaries(
            self.posterior.params,
            {"variational_family": self.variational_family.params},
        )
        return hyperparams

    @abc.abstractmethod
    def mean(self, train_data: Dataset, params: dict) -> Callable[[Dataset], Array]:
        raise NotImplementedError

    @abc.abstractmethod
    def variance(self, train_data: Dataset, params: dict) -> Callable[[Dataset], Array]:
        raise NotImplementedError

    @abc.abstractmethod
    def elbo(
        self, train_data: Dataset, transformations: Dict
    ) -> Callable[[Array], Array]:
        raise NotImplementedError


@dataclass
class SVGP(VariationalPosterior):
    def __post_init__(self):
        self.prior = self.posterior.prior
        self.likelihood = self.posterior.likelihood
        self.num_inducing = self.variational_family.num_inducing

    def elbo(
        self, train_data: Dataset, transformations: Dict, negative: bool = False
    ) -> Callable[[Array], Array]:
        constant = jnp.array(-1.0) if negative else jnp.array(1.0)

        def elbo_fn(params: Dict, batch: Dataset) -> Array:
            params = transform(params, transformations)
            kl = self.prior_kl(params)
            var_exp = self.variational_expectation(params, batch)

            return constant * (jnp.sum(var_exp) * train_data.n / batch.n - kl)

        return elbo_fn

    # Compute KL divergence at inducing points, KL[q(u)||p(u)]:
    def prior_kl(self, params: Dict) -> Array:
        mu = params["variational_family"]["variational_mean"]
        sqrt = params["variational_family"]["variational_root_covariance"]
        nz = self.num_inducing

        qu = dx.MultivariateNormalTri(mu.squeeze(), sqrt)

        if not self.variational_family.whiten:
            z = params["variational_family"]["inducing_inputs"]
            mz = self.prior.mean_function(z, params["mean_function"])
            Kzz = gram(self.prior.kernel, z, params["kernel"])
            Kzz += I(nz) * self.jitter
            Lz = cholesky(Kzz)
            pu = dx.MultivariateNormalTri(mz.squeeze(), Lz)

        else:
            pu = dx.MultivariateNormalDiag(jnp.zeros(nz))

        return qu.kl_divergence(pu)

    def variational_expectation(self, params: Dict, batch: Dataset) -> Array:
        x, y = batch.X, batch.y

        Fmu, Fvar = vmap(self.pred_moments, in_axes=(None, 0))(
            params, x[:, jnp.newaxis, :]
        )

        # Get log(p(y|F)) function for current likelihood parameter values:
        def log_prob(F, y):
            return self.likelihood.link_function(F, params["likelihood"]).log_prob(y)

        return gauss_hermite_quadrature(log_prob, Fmu.squeeze(1), Fvar.squeeze(1), y=y)

    # Computes predictive moments for Gauss-Hermite quadrature:
    def pred_moments(self, params: Dict, test_inputs: Array) -> Tuple[Array, Array]:
        mu = params["variational_family"]["variational_mean"]
        sqrt = params["variational_family"]["variational_root_covariance"]

        # Cholesky decomposition at inducing inputs:
        z = params["variational_family"]["inducing_inputs"]
        nz = self.num_inducing
        Kzz = gram(self.prior.kernel, z, params["kernel"])
        Kzz += I(nz) * self.jitter
        Lz = cholesky(Kzz)

        # Compute predictive moments:
        t = test_inputs
        Ktt = gram(self.prior.kernel, t, params["kernel"])
        Kzt = cross_covariance(self.prior.kernel, z, t, params["kernel"])
        M = solve_triangular(Lz, Kzt, lower=True)
        Fcov = Ktt - jnp.matmul(M.T, M)

        if not self.variational_family.whiten:
            M = solve_triangular(Lz.T, M, lower=False)
            mz = self.prior.mean(params)(z)
            mu -= mz

        Fmu = self.prior.mean(params)(t) + jnp.matmul(M.T, mu)
        V = jnp.matmul(M.T, sqrt)
        Fcov += jnp.matmul(V, V.T)

        return Fmu, Fcov

    def mean(self, params: Dict) -> Callable[[Array], Array]:
        # Cholesky decomposition at inducing inputs:
        z = params["variational_family"]["inducing_inputs"]
        nz = self.num_inducing
        Kzz = gram(self.prior.kernel, z, params["kernel"])
        Kzz += I(nz) * self.jitter
        Lz = cholesky(Kzz)

        # Variational mean:
        mu = params["variational_family"]["variational_mean"]
        if not self.variational_family.whiten:
            mz = self.prior.mean(params)(z)
            mu -= mz

        def mean_fn(test_inputs: Array):
            t = test_inputs
            Kzt = cross_covariance(self.prior.kernel, z, t, params["kernel"])
            M = solve_triangular(Lz, Kzt, lower=True)

            if not self.variational_family.whiten:
                M = solve_triangular(Lz.T, M, lower=False)

            mt = self.prior.mean(params)(t)

            return mt + jnp.matmul(M.T, mu)

        return mean_fn

    def variance(self, params: Dict) -> Callable[[Array], Array]:
        # Cholesky decomposition at inducing inputs:
        z = params["variational_family"]["inducing_inputs"]
        nz = self.num_inducing
        Kzz = gram(self.prior.kernel, z, params["kernel"])
        Kzz += I(nz) * self.jitter
        Lz = cholesky(Kzz)

        # Variational sqrt cov:
        sqrt = params["variational_family"]["variational_root_covariance"]

        def variance_fn(test_inputs: Array):
            t = test_inputs
            Ktt = gram(self.prior.kernel, t, params["kernel"])
            Kzt = cross_covariance(self.prior.kernel, z, t, params["kernel"])
            M = solve_triangular(Lz, Kzt, lower=True)
            Fcov = Ktt - jnp.matmul(M.T, M)

            if not self.variational_family.whiten:
                M = solve_triangular(Lz.T, M, lower=False)

            V = jnp.matmul(M.T, sqrt)
            Fcov += jnp.matmul(V, V.T)

            return Fcov

        return variance_fn
